package com.wanji.indicators.event.cross.accident;

import com.alibaba.fastjson.JSONObject;
import com.wanji.indicators.event.cross.accident.funcnew.*;
import com.wanji.indicators.model.CrossFrameModel;
import com.wanji.indicators.model.event.accident.AccidentFrameModel;
import com.wanji.indicators.model.event.accident.AccidentShowModel;
import com.wanji.indicators.sink.AccidentEventFileSink;
import com.wanji.indicators.sink.EventFileSink;
import com.wanji.indicators.source.ConflictTurnSource;
import com.wanji.indicators.source.ConsulAreaSource;
import com.wanji.indicators.util.PropertiesHelper;
import org.apache.flink.api.common.functions.FlatMapFunction;
import org.apache.flink.api.common.serialization.SimpleStringSchema;
import org.apache.flink.streaming.api.datastream.DataStream;
import org.apache.flink.streaming.api.datastream.SingleOutputStreamOperator;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.connectors.kafka.FlinkKafkaProducer;
import org.apache.flink.streaming.connectors.kafka.internals.KeyedSerializationSchemaWrapper;
import org.apache.flink.util.Collector;

import java.io.Serializable;
import java.util.Properties;

/**
 * 疑似异常停车/事故计算
 *
 * @author fengyi
 */
public class AccidentEventMainNew implements Serializable {

    private final StreamExecutionEnvironment env;
    private final PropertiesHelper instance;
    private final Properties p;
    private String path;
    //本地调试设为true输出结果保存到文件
    private boolean isDebug = false;

    private AccidentEventMainNew(StreamExecutionEnvironment env, String path, boolean isDebug) {
        this.env = env;
        this.instance = PropertiesHelper.getInstance();
        this.p = instance.getProperties();
        this.path = path;
        this.isDebug = isDebug;
    }
    public static AccidentEventMainNew init(StreamExecutionEnvironment env, String path, boolean isDebug) {
        return new AccidentEventMainNew(env, path,isDebug);
    }

    private AccidentEventMainNew(StreamExecutionEnvironment env) {
        this.env = env;
        this.instance = PropertiesHelper.getInstance();
        this.p = instance.getProperties();
    }
    private AccidentEventMainNew(StreamExecutionEnvironment env, String path) {
        this.env = env;
        this.instance = PropertiesHelper.getInstance();
        this.p = instance.getProperties();
        this.path = path;
    }

    public static AccidentEventMainNew init(StreamExecutionEnvironment env) {
        return new AccidentEventMainNew(env);
    }

    public static AccidentEventMainNew init(StreamExecutionEnvironment env, String path) {
        return new AccidentEventMainNew(env, path);
    }
    /**
     * 开始计算
     */
    public void run(DataStream<CrossFrameModel> thinningDataStream) {

        //初始化kafka生产者
        FlinkKafkaProducer producer = new FlinkKafkaProducer<>(p.getProperty("cross.event.data.topic"),
                new KeyedSerializationSchemaWrapper<>(new SimpleStringSchema()),
                instance.getProducerProperties(), FlinkKafkaProducer.Semantic.NONE);

        //解析数据
        SingleOutputStreamOperator<AccidentFrameModel> accidentFrameModelFilter = thinningDataStream
                .flatMap(new AccidentFlatMap())
                .setParallelism(1)
                .name("事故检测-解析数据");

        //匹配拥堵检测区域
//        SingleOutputStreamOperator<AccidentFrameModel> checkCongestionArea = accidentFrameModelFilter
//                .connect(env.addSource(new ConsulAreaSource()).broadcast())
//                .flatMap(new AccidentCheckCongestionAreaCoFlatMap())
//                .setParallelism(1)
//                .name("事故检测-匹配拥堵检测区域");

        //加载冲突相位信息
        SingleOutputStreamOperator<AccidentFrameModel> loadConflictSourceProcess = accidentFrameModelFilter
                .connect(env.addSource(new ConflictTurnSource()).broadcast())
                .flatMap(new AbnormalDeadlockConflictSlotCoFlatMap())
                .setParallelism(1)
                .name("事故检测-加载冲突相位信息");

        //路口冲突检测
        SingleOutputStreamOperator<AccidentFrameModel> conflictDirectProcess = loadConflictSourceProcess
                .keyBy(AccidentFrameModel::getCrossId)
                .process(new AbnormalDeadlockConflictCheckKeyedProcess())
                .setParallelism(1)
                .name("事故检测-路口冲突转向车辆检测");

        //计算停驶车辆和停车持续时间
        SingleOutputStreamOperator<AccidentFrameModel> checkStopKeyedProcess = conflictDirectProcess
                .keyBy(AccidentFrameModel::getCrossId)
                .process(new AccidentStopCheckKeyedProcess())
                .setParallelism(1)
                .name("事故检测-计算停驶车辆和停车持续时间");



        //计算是否有同向行驶的车辆
        SingleOutputStreamOperator<AccidentFrameModel> checkSameDirKeyedProcess = checkStopKeyedProcess
                .keyBy(AccidentFrameModel::getCrossId)
                .process(new AccidentSameDriDriveKeyedProcess())
                .setParallelism(1)
                .name("事故检测-检测停车车辆是否存在相同转向运行车辆");


        //根据事故持续时间计算事故结果
        SingleOutputStreamOperator<AccidentFrameModel> accidentFrameFlatMap = checkSameDirKeyedProcess
                .flatMap(new AccidentCheckTimeFlatMap())
                .setParallelism(1)
                .name("事故检测-根据事故持续时间计算事故结果");

        //事故逻辑去重
        SingleOutputStreamOperator<AccidentFrameModel> duplicateRemoveKeyedProcess = accidentFrameFlatMap
                .keyBy(AccidentFrameModel::getCrossId)
                .process(new AccidentDuplicateRemoveKeyedProcess())
                .setParallelism(1)
                .name("事故检测-计算事故结果逻辑去重");

        //事故辅助定责
        SingleOutputStreamOperator<AccidentShowModel> accidentShowModel = duplicateRemoveKeyedProcess
                .flatMap(new AccidentEncapsulationFlatMap())
                .setParallelism(1)
                .name("事故检测-计算事故辅助定责");

        //事故数据转换
        SingleOutputStreamOperator<String> accidentOutputStream = accidentShowModel
                .flatMap(new FlatMapFunction<AccidentShowModel, String>() {
                    @Override
                    public void flatMap(AccidentShowModel model, Collector<String> out) {
                        model.setTime(model.getStartTime());
                        out.collect(JSONObject.toJSONString(model));
                    }
                }).name("事故检测-数据格式转换结果封装")
                .setParallelism(1);

        if (!isDebug) {
            accidentOutputStream.addSink(producer).name("事故计算-数据发送至kafka").setParallelism(1);
            accidentOutputStream.print("事故计算-结果打印").name("事故计算-结果打印");
        } else {
            accidentShowModel.addSink(new AccidentEventFileSink(path)).setParallelism(1).name("事故计算-结果输出至文件");
        }

    }



//    /**
//     * 事故持续时间/影响程度计算
//     */
//    private void accidentLastsFor(SingleOutputStreamOperator<EventCross> eventCrossStream) {
//
//        SingleOutputStreamOperator<AccidentTimingModel> accidentTimingModel = eventCrossStream
//                .flatMap(new AccidentInfluenceDegreeFlatMap())
//                .setParallelism(1)
//                .name("事故计算-持续时间/交通影响严重程度");
//
//        SingleOutputStreamOperator<AccidentTimingModel> reduceAccidentTiming = accidentTimingModel
//                .keyBy(AccidentTimingModel::getId)
//                .window(TumblingProcessingTimeWindows.of(Time.seconds(1)))
//                .reduce(new AccidentTimingReduce())
//                .setParallelism(1)
//                .name("事故持续时间+交通影响严重程度聚合-开窗1秒聚合");
//
////        reduceAccidentTiming.addSink(new AccidentDurationRedisSink(instance))
////                .setParallelism(1)
////                .name("事故持续时间数据输出至redis");
//
//        SingleOutputStreamOperator<AccidentTimingModel> result = reduceAccidentTiming
//                .keyBy(AccidentTimingModel::getId)
//                .window(ProcessingTimeSessionWindows.withGap(Time.seconds(30)))
//                .reduce(new ReduceFunction<AccidentTimingModel>() {
//                    @Override
//                    public AccidentTimingModel reduce(AccidentTimingModel a, AccidentTimingModel b) throws Exception {
//                        return b;
//                    }
//                }).setParallelism(1)
//                .name("事故消散状态计算");
//
//        SingleOutputStreamOperator<String> accidentTimingToES = result.flatMap(new FlatMapFunction<AccidentTimingModel, String>() {
//            @Override
//            public void flatMap(AccidentTimingModel model, Collector<String> out) {
//                String id = model.getId();
//                model.setId(id.replace("duration_accident_info:", Constant.EMPTY_STRING));
//                String s = JSONObject.toJSONString(model);
//                out.collect(s);
//            }
//        }).setParallelism(1).name("事故状态结果集封装");
//
//        FlinkKafkaProducer producer = new FlinkKafkaProducer<>("cross.event.data.accident.time",
//                new KeyedSerializationSchemaWrapper<>(new SimpleStringSchema()),
//                instance.getProducerProperties(), FlinkKafkaProducer.Semantic.NONE);
//
//        accidentTimingToES.addSink(producer).setParallelism(1).name("事故持续时间数据输出至kafka");
//        accidentTimingToES.print("事故持续时间-结果打印").name("事故持续时间-结果打印");
//
//    }

}
