package com.wanji.indicators.task.export;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.wanji.indicators.model.RoutePathCarTrackModel;
import com.wanji.indicators.task.track.service.sink.FileSink;
import com.wanji.indicators.util.DateUtil;
import com.wanji.indicators.util.PropertiesHelper;
import org.apache.flink.api.common.JobID;
import org.apache.flink.api.common.eventtime.WatermarkStrategy;
import org.apache.flink.api.common.functions.FilterFunction;
import org.apache.flink.api.common.functions.FlatMapFunction;
import org.apache.flink.api.common.state.MapState;
import org.apache.flink.api.common.state.MapStateDescriptor;
import org.apache.flink.api.common.state.StateTtlConfig;
import org.apache.flink.api.common.time.Time;
import org.apache.flink.api.common.typeinfo.TypeHint;
import org.apache.flink.api.common.typeinfo.TypeInformation;
import org.apache.flink.api.java.tuple.Tuple;
import org.apache.flink.api.java.tuple.Tuple3;
import org.apache.flink.api.java.utils.ParameterTool;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.connector.kafka.source.KafkaSource;
import org.apache.flink.connector.kafka.source.enumerator.initializer.OffsetsInitializer;
import org.apache.flink.connector.kafka.source.reader.deserializer.KafkaRecordDeserializationSchema;
import org.apache.flink.core.execution.JobClient;
import org.apache.flink.core.execution.JobListener;
import org.apache.flink.streaming.api.datastream.DataStream;
import org.apache.flink.streaming.api.datastream.SingleOutputStreamOperator;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.api.functions.KeyedProcessFunction;
import org.apache.flink.streaming.api.functions.windowing.ProcessWindowFunction;
import org.apache.flink.streaming.api.windowing.assigners.ProcessingTimeSessionWindows;
import org.apache.flink.streaming.api.windowing.windows.TimeWindow;
import org.apache.flink.util.Collector;
import org.apache.kafka.common.record.TimestampType;
import org.apache.kafka.common.serialization.StringDeserializer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;

public class TrackExportToFileMain {
    private static final Logger log = LoggerFactory.getLogger(TrackExportToFileMain.class);


    public static void main(String[] args) {
        //批量轨迹es index,逗号分割：esIndexName,windowTime


        ParameterTool parameter = ParameterTool.fromArgs(args);
        String bootstrapServers = parameter.get("bootstrapServers");
        String topic = parameter.get("topic");
        String startTime = parameter.get("startTime");
        String endTime = parameter.get("endTime");
        String sinkPath = parameter.get("sinkPath");
        //目录
        if (!sinkPath.endsWith("/")) {
            sinkPath += "/";
        }
        sinkPath += topic+"_"+startTime + "-" + endTime + ".json";
        log.info("导出路径：" + sinkPath);

        long startTimestamp = DateUtil.StringToMillis(startTime, "yyyy_MM_dd_HH_mm_ss");
        long endTimestamp = DateUtil.StringToMillis(endTime, "yyyy_MM_dd_HH_mm_ss");

        PropertiesHelper instance = PropertiesHelper.getInstance();
        Properties properties = instance.getProperties();

        StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();

        KafkaSource<KafkaRecordModel> source = KafkaSource.<KafkaRecordModel>builder()
                .setProperties(instance.getConsumerProperties())
                .setProperty("auto.offset.commit", "true")
                .setProperty("auto.commit.interval.ms", "1000")
                .setProperty("commit.offsets.on.checkpoint", "true")
                .setBootstrapServers(bootstrapServers)
                .setTopics(topic)
                .setGroupId(properties.getProperty("consumer.group.id") + "-Export")
                // 从大于等于此时间戳开始的偏移量开始
                .setStartingOffsets(OffsetsInitializer.timestamp(startTimestamp))

                .setDeserializer(KafkaRecordDeserializationSchema.of(new ReadKafkaDeserialization()))

                .build();

        DataStream<KafkaRecordModel> stream = env
                .fromSource(source, WatermarkStrategy.noWatermarks(), "kafka-car-track-source-export");


        SingleOutputStreamOperator<KafkaRecordModel> frameModelStream =
                stream.filter(new FilterFunction<KafkaRecordModel>() {
                    @Override
                    public boolean filter(KafkaRecordModel value) throws Exception {
                        Long ts = value.getTimestamp();
                        //Long endTs = DateUtil.StringToMillis("2023-03-30 11:40:00:999","yyyy-MM-dd HH:mm:ss:SSS");
                        if (ts <= endTimestamp) {
                            return true;
                        }
                        return false;
                    }
                }).name("过滤出时间范围内数据");

        SingleOutputStreamOperator<KafkaRecordModel> timeOutStream = frameModelStream
                .keyBy(KafkaRecordModel::getTopic)
                .process(new KeyedProcessFunction<String, KafkaRecordModel, KafkaRecordModel>() {
                    private transient MapState<String, Long> dataTimeState;
                    //连续点间隔超时时间，单位分钟
                    private Integer timeOut = 10;

                    @Override
                    public void open(Configuration parameters) throws Exception {
                        StateTtlConfig ttlConfig = StateTtlConfig.newBuilder(Time.hours(12))//状态过期时间
                                .setUpdateType(StateTtlConfig.UpdateType.OnReadAndWrite)
                                .cleanupFullSnapshot()
                                .setStateVisibility(StateTtlConfig.StateVisibility.ReturnExpiredIfNotCleanedUp)//不返回过期值
                                .build();

                        MapStateDescriptor mapStateDescriptor = new MapStateDescriptor<>("Kafka_Data_Export",
                                TypeInformation.of(new TypeHint<Tuple3<String, Integer, String>>() {
                                }),
                                TypeInformation.of(new TypeHint<RoutePathCarTrackModel>() {
                                })
                        );
                        mapStateDescriptor.enableTimeToLive(ttlConfig);
                        dataTimeState = getRuntimeContext().getMapState(mapStateDescriptor);
                    }

                    @Override
                    public void processElement(KafkaRecordModel value, Context context, Collector<KafkaRecordModel> out) throws Exception {
                        out.collect(value);

                        int interval = 60 * 1000;
                        long fireTime = context.timerService().currentProcessingTime() + interval;
                        context.timerService().registerProcessingTimeTimer(fireTime);

                        dataTimeState.put(value.getTopic(), context.timerService().currentProcessingTime());

                        //log.info("register process time:" + DateUtil.toDateTime(fireTime, "yyyy-MM-dd HH:mm:ss"));
                    }

                    @Override
                    public void onTimer(long timestamp, OnTimerContext ctx, Collector<KafkaRecordModel> out) throws Exception {

                        Iterator<Map.Entry<String, Long>> iterator = dataTimeState.iterator();
                        while (iterator.hasNext()) {
                            Map.Entry<String, Long> entry = iterator.next();
                            long dataTime = entry.getValue();
                            //log.info("dataTime:"+DateUtil.toDateTime(dataTime,"yyyy-MM-dd HH:mm:ss")+" onTime:" + DateUtil.toDateTime(timestamp, "yyyy-MM-dd HH:mm:ss"));

                            if (timestamp - dataTime > 30 * 1000) {
                                //超时未收到新数据
                                ctx.timerService().deleteProcessingTimeTimer(timestamp);
                               // env.executeAsync().cancel();
                                //System.exit(0);
                            }
                        }
                    }
                });

        SingleOutputStreamOperator<String> toJsonStream =
                timeOutStream.flatMap(
                        new FlatMapFunction<KafkaRecordModel, String>() {
                            @Override
                            public void flatMap(KafkaRecordModel value, Collector<String> out) throws Exception {
                                Long ts = value.getTimestamp();
                                //Long endTs = DateUtil.StringToMillis("2023-03-30 11:40:00:999","yyyy-MM-dd HH:mm:ss:SSS");
//                                                if (ts > endTimestamp) {
//                                                    System.exit(0);
//                                                }
                                log.info("记录时间：{},数据导出范围【{},{}】", DateUtil.toDateTime(ts, "yyyy-MM-dd HH:mm:ss.SSS"), startTime, endTime);
                                String json = JSON.toJSONString(value);
                                out.collect(json);

                            }
                        })
                        .name("转Json输出");


        try {

            toJsonStream.addSink(new FileSink(sinkPath));

            env.execute("根据起始时间范围导出数据到文件");
            //env.executeAsync().cancel();
        } catch (Exception e) {
            e.printStackTrace();
            //log.error("交通指标计算任务异常 :   " + e);
        }
    }
}
